"""This module includes the commands to import an existing dataset. 
PyLabel current supports importing labels from COCO, YOLO, and VOC formats. 
You can also import set of images that do not have labels yet and label them manually using the PyLabel
labelling tool. """

import json
import pandas as pd
import xml.etree.ElementTree as ET
import os
from os.path import exists
from pathlib import Path, PurePath
import copy
import cv2
import yaml

from pylabel.shared import schema
from pylabel.dataset import Dataset
from pylabel.exporter import Export


def _GetValueOrBlank(element, user_input=None):
    """
    If an element is missing from the XML file reading the .text value will return an error.
    If the element does not exist return ""
    """
    if user_input == None:
        return element.text
    else:
        return user_input


# These are the valid columns in the pylabel annotations table.
def ImportCoco(path, path_to_images=None, name=None):
    """
    This function takes the path to a JSON file in COCO format as input. It returns a PyLabel dataset object that contains the annotations.

    Returns:
        PyLabel dataset object.

    Args:
        path (str):The path to the JSON file with the COCO annotations.
        path_to_images (str): The path to the images relative to the json file.
            If the images are in the same directory as the JSON file then omit this parameter.
            If the images are in a different directory on the same level as the annotations then you would
            set `path_to_images='../images/'`
        name (str): This will set the dataset.name property for this dataset.
            If not specified, the filename (without extension) of the COCO annotation file file will be used as the dataset name.
    Example:
        >>> from pylabel import importer
        >>> dataset = importer.ImportCoco("coco_annotations.json")
    """
    with open(path) as cocojson:
        annotations_json = json.load(cocojson)

    # Store the 3 sections of the json as seperate json arrays
    images = pd.json_normalize(annotations_json["images"])
    images.columns = "img_" + images.columns
    try:
        images["img_folder"]
    except:
        images["img_folder"] = ""
    # print(images)

    # If the user has specified a different image folder then use that one
    if path_to_images != None:
        images["img_folder"] = path_to_images

    astype_dict = {"img_width": "int64", "img_height": "int64", "img_depth": "int64"}
    astype_keys = list(astype_dict.keys())
    for element in astype_keys:
        if element not in images.columns:
            astype_dict.pop(element)
    # print(astype_dict)
    # images = images.astype({'img_width': 'int64','img_height': 'int64','img_depth': 'int64'})
    images = images.astype(astype_dict)

    annotations = pd.json_normalize(annotations_json["annotations"])
    annotations.columns = "ann_" + annotations.columns

    categories = pd.json_normalize(annotations_json["categories"])
    categories.columns = "cat_" + categories.columns
    
    # Converting this to string resolves issue #23
    categories.cat_id = categories.cat_id.astype(str)

    df = annotations

    # Converting this to string resolves issue #23
    df.ann_category_id = df.ann_category_id.astype(str)
    
    df[
        ["ann_bbox_xmin", "ann_bbox_ymin", "ann_bbox_width", "ann_bbox_height"]
    ] = pd.DataFrame(df.ann_bbox.tolist(), index=df.index)
    df.insert(8, "ann_bbox_xmax", df["ann_bbox_xmin"] + df["ann_bbox_width"])
    df.insert(10, "ann_bbox_ymax", df["ann_bbox_ymin"] + df["ann_bbox_height"])

    # debug print(df.info())

    # Join the annotions with the information about the image to add the image columns to the dataframe
    df = pd.merge(images, df, left_on="img_id", right_on="ann_image_id", how="left")
    df = pd.merge(
        df, categories, left_on="ann_category_id", right_on="cat_id", how="left"
    )

    # Rename columns if needed from the coco column name to the pylabel column name
    df.rename(columns={"img_file_name": "img_filename"}, inplace=True)

    # Drop columns that are not in the schema
    df = df[df.columns.intersection(schema)]

    # Add missing columns that are in the schema but not part of the table
    df[list(set(schema) - set(df.columns))] = ""

    # Reorder columns
    df = df[schema]
    df.index.name = "id"
    df.annotated = 1

    # These should be strings
    df.cat_id = df.cat_id.astype(str)

    # These should be integers
    df.img_width = df.img_width.astype(int)
    df.img_height = df.img_height.astype(int)

    dataset = Dataset(df)

    # Assign the filename (without extension) as the name of the dataset
    if name == None:
        dataset.name = Path(path).stem
    else:
        dataset.name = name

    dataset.path_to_annotations = PurePath(path).parent

    return dataset


def ImportVOC(path, path_to_images=None, name="dataset"):
    """
    Provide the path a directory with annotations in VOC Pascal XML format and it returns a PyLabel dataset object that contains the annotations.

    Returns:
        PyLabel dataset object.

    Args:
        path (str): The path to the directory with the annotations in VOC Pascal XML format.
        path_to_images (str): The path to the images relative to the annotations.
            If the images are in the same directory as the annotation files then omit this parameter.
            If the images are in a different directory on the same level as the annotations then you would
            set `path_to_images='../images/'`
        name (str): Default is 'dataset'. This will set the dataset.name property for this dataset.

    Example:
        >>> from pylabel import importer
        >>> dataset = importer.ImportVOC(path="annotations/", path_to_images="../images/")
    """
    # Create an empty dataframe
    df = pd.DataFrame(columns=schema)

    # the dictionary to pass to pandas dataframe
    d = {}

    row_id = 0
    img_id = 0
    cat_names = []

    def GetCatId(cat_name):
        """This will assign a numeric cat_id to each cat_name."""
        if cat_name not in cat_names:
            cat_names.append(cat_name)
        return cat_names.index(cat_name)

    # iterate over files in that directory
    for filename in os.scandir(path):
        if filename.is_file() and filename.name.endswith(".xml"):
            filepath = filename.path
            xml_data = open(filepath, "r").read()  # Read file
            root = ET.XML(xml_data)  # Parse XML
            folder = _GetValueOrBlank(root.find("folder"), user_input=path_to_images)
            filename = root.find("filename").text
            size = root.find("size")
            size_width = size.find("width").text
            size_height = size.find("height").text
            size_depth = _GetValueOrBlank(size.find("depth"))
            segmented = _GetValueOrBlank(root.find("segmented"))

            row = {}
            # Build dictionary that will be become the row in the dataframe
            row["img_folder"] = folder
            row["img_filename"] = filename
            row["img_id"] = img_id
            row["img_width"] = size_width
            row["img_height"] = size_height
            row["img_depth"] = size_depth
            row["ann_segmented"] = segmented

            object = root.findall("object")

            for o in object:
                row["cat_name"] = o.find("name").text
                row["cat_id"] = GetCatId(row["cat_name"])
                row["ann_pose"] = _GetValueOrBlank(o.find("pose"))
                row["ann_truncated"] = _GetValueOrBlank(o.find("truncated"))
                row["ann_difficult"] = _GetValueOrBlank(o.find("difficult"))
                row["ann_bbox_xmin"] = float(o.find("bndbox").find("xmin").text)
                row["ann_bbox_ymin"] = float(o.find("bndbox").find("ymin").text)
                row["ann_bbox_xmax"] = float(o.find("bndbox").find("xmax").text)
                row["ann_bbox_ymax"] = float(o.find("bndbox").find("ymax").text)
                row["ann_bbox_width"] = row["ann_bbox_xmax"] - row["ann_bbox_xmin"]
                row["ann_bbox_height"] = row["ann_bbox_ymax"] - row["ann_bbox_ymin"]
                row["ann_area"] = row["ann_bbox_width"] * row["ann_bbox_height"]
                row["split"] = ""

                # Add this row to the dict
                d[row_id] = copy.deepcopy(row)
                # increment the rowid
                row_id += 1

        # Increment the imageid because we are going to read annother file
        img_id += 1

    # Convert the dict with all of the annotation data to a dataframe
    df = pd.DataFrame.from_dict(d, "index", columns=schema)
    df.index.name = "id"
    df.annotated = 1

    # These should be strings
    df.cat_id = df.cat_id.astype(str)

    # These should be integers
    df.img_width = df.img_width.astype(int)
    df.img_height = df.img_height.astype(int)

    # Reorder columns
    df = df[schema]
    dataset = Dataset(df)
    dataset.name = name
    dataset.path_to_annotations = path
    # Get the path without the filename
    # dataset.path_to_annotations = "Alex ander"
    return dataset


def ImportYoloV5(
    path,
    img_ext="jpg",
    cat_names=[],
    path_to_images="",
    name="dataset",
):
    """
    Provide the path a directory with annotations in YOLO format and it returns a PyLabel dataset object that contains the annotations.
    The Yolo format does not store much information about the images, such as the height and width. When you import a
    Yolo dataset PyLabel will extract this information from the images.

    Returns:
        PyLabel dataset object.

    Args:
        path (str): The path to the directory with the annotations in YOLO format.
        img_ext (str): Specify the file extension of your images: .jpeg, .png, etc. Because the YOLO format
            does not store filename of the image, it must be specified if you are converting from YOLO to another format.
        cat_names (list): YOLO anotations only store a class number, not the name. You can provide a list of class ids
            that correspend to the int used to represent that class in the annotations. For example `['Squirrel,'Nut']`.

            If you have the class names already stored in a YOLO YAML file then use the ImportYoloV5WithYaml method to
            automatically read the class names from that file.
        path_to_images (str): The path to the images relative to the annotations.
            If the images are in the same directory as the annotation files then omit this parameter.
            If the images are in a different directory on the same level as the annotations then you would
            set `path_to_images='../images/'`
        name (str): Default is 'dataset'. This will set the dataset.name property for this dataset.

    Example:
        >>> from pylabel import importer
        >>> dataset = importer.ImportYoloV5(path="labels/", path_to_images="../images/")
    """

    def GetCatNameFromId(cat_id, cat_names):
        cat_id = int(cat_id)
        if len(cat_names) > int(cat_id):
            return cat_names[cat_id]

    # Create an empty dataframe
    df = pd.DataFrame(columns=schema)

    # the dictionary to pass to pandas dataframe
    d = {}

    row_id = 0
    img_id = 0

    # iterate over files in that directory
    for filename in os.scandir(path):
        if filename.is_file() and filename.name.endswith(".txt"):
            filepath = filename.path
            file = open(filepath, "r")  # Read file

            for line in file:
                row = {}

                (
                    cat_id,
                    x_center_norm,
                    y_center_norm,
                    width_norm,
                    height_norm,
                ) = line.split()
                row["img_folder"] = path_to_images
                row["img_filename"] = filename.name.replace("txt", img_ext)

                # Get the path to the image file to extract the height, width, and depth
                image_path = PurePath(path, path_to_images, row["img_filename"])

                # Check if there is a file at this location.
                assert exists(
                    image_path
                ), f"File does not exist: {image_path}. Check path_to_images and img_ext arguments."

                im = cv2.imread(str(image_path))
                img_height, img_width, img_depth = im.shape

                row["img_id"] = img_id
                row["img_width"] = img_width
                row["img_height"] = img_height
                row["img_depth"] = img_depth

                row["ann_bbox_width"] = float(width_norm) * img_width
                row["ann_bbox_height"] = float(height_norm) * img_height
                row["ann_bbox_xmin"] = float(x_center_norm) * img_width - (
                    (row["ann_bbox_width"] / 2)
                )
                row["ann_bbox_ymax"] = float(y_center_norm) * img_height + (
                    (row["ann_bbox_height"] / 2)
                )
                row["ann_bbox_xmax"] = row["ann_bbox_xmin"] + row["ann_bbox_width"]
                row["ann_bbox_ymin"] = row["ann_bbox_ymax"] - row["ann_bbox_height"]

                row["ann_area"] = row["ann_bbox_width"] * row["ann_bbox_height"]

                row["cat_id"] = cat_id
                row["cat_name"] = GetCatNameFromId(cat_id, cat_names)

                # Add this row to the dict
                d[row_id] = row
                row_id += 1
        img_id += 1

    df = pd.DataFrame.from_dict(d, "index", columns=schema)
    df.index.name = "id"
    df.annotated = 1

    # These should be strings
    df.cat_id = df.cat_id.astype(str)

    # These should be integers
    df.img_width = df.img_width.astype(int)
    df.img_height = df.img_height.astype(int)

    # Reorder columns
    dataset = Dataset(df)
    dataset.name = name
    dataset.path_to_annotations = path

    return dataset


def ImportImagesOnly(path, name="dataset"):
    """Import a directory of images as a dataset with no annotations.
    Then use PyLabel to annote the images. Will import images with these extensions:
    ('.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif')

    Args:
        path (str):
            The path to the directory with the images.
        name (str):
            Default is 'dataset'. Descriptive name, which is used when outputting files.
    Returns:
        A dataset object with one row for each image and no annotations.

    Example:
        >>> from pylabel import importer
        >>> dataset = importer.ImportImagesOnly(path="images/")
    """

    # Create an empty dataframe
    df = pd.DataFrame(columns=schema)

    # the dictionary to pass to pandas dataframe
    d = {}

    img_id = 0

    # iterate over files in that directory
    for filename in os.scandir(path):
        if filename.is_file() and filename.name.lower().endswith(
            (".png", ".jpg", ".jpeg", ".tiff", ".bmp", ".gif")
        ):
            row = {}
            row["img_folder"] = ""
            row["img_filename"] = filename.name
            image_path = PurePath(path, row["img_filename"])
            im = cv2.imread(str(image_path))

            try:
                # If the file is not an image then this will fail
                im.shape
            except:
                raise ValueError(
                    f"Error reading file '{image_path}'. Exclude non-image files by using the ends_width param."
                )

            img_height, img_width, img_depth = im.shape

            row["img_id"] = img_id
            row["img_width"] = img_width
            row["img_height"] = img_height
            row["img_depth"] = img_depth
            row["cat_name"] = ""

            # Add this row to the dict
            d[img_id] = row
            img_id += 1

    df = pd.DataFrame.from_dict(d, "index", columns=schema)
    df.index.name = "id"

    # Reorder columns
    dataset = Dataset(df)
    dataset.name = name
    dataset.path_to_annotations = path

    return dataset


def _yaml_reader(yaml_file):
    """Import the YAML File for the YOLOv5 data as dict."""
    with open(yaml_file) as file:
        data = yaml.safe_load(file)
    return data


def ImportYoloV5WithYaml(
    yaml_file,
    image_ext="jpg",
    name_of_annotations_folder="labels",
    path_to_annotations=None,
):
    """Import a YOLO dataset by reading the YAML file to extract the class names, image and label locations,
    and preserve if an image should be in the train, test, or val split.

    Returns:
        PyLabel dataset object.

    Args:
        yaml_file (str):
            Path to the yaml file that describes the dataset to be imported.
        image_ext (str):
            The image file extension.
        path_to_annotations (str):
            the path to the annotations file; if path to annotations is none, file replaces name of images file from yaml file with annotations.
        name_of_annotations_folder (str):
            Default is "labels". Change this to "annotations" if your folder is called "annotations"

    Example:
        >>> from pylabel import importer
        >>> dataset = importer.ImportYoloV5WithYaml(yaml_file='data/dataset.yaml')

    """

    """
    Note to other developers:
    As a note, the "path_to_images" variable in this code refers to the relative path relative to the path to annotations.
    It is different than the path to images specified in the YAML file.
    PyLabel uses the former to establish its pathing and the latter path to actually view the needed data.
    """

    path_to_annotations_copy = path_to_annotations

    if path_to_annotations == None:
        path_to_annotations_defined = False
    else:
        path_to_annotations_defined = True
    counter = 0

    data = _yaml_reader(yaml_file)
    yoloclasses = data["names"]

    iterated_list = list(data.keys())

    for splitted in iterated_list:
        if splitted in ["nc", "names"]:
            pass
        else:
            try:
                path_to_images = data[splitted]
            except:
                raise Exception("split type not in the YAML file.")
            # if counter > 0
            # change PoA to new split type

            # if
            if path_to_annotations == None or counter != 0:
                # In case your folder is called labels or some thing else that doesn't jive with what we want you to call it.
                if path_to_annotations_defined == True and counter > 0:
                    path_to_annotations = str(
                        PurePath(
                            path_to_annotations_copy.replace(
                                iterated_list[counter - 1], splitted
                            )
                        )
                    )

                # This probably needs to be reworked but there's potentially an issue with resetting the path
                # to annotations as we iterate through each split type.
                elif name_of_annotations_folder != "labels" and counter > 0:
                    path_to_annotations = str(
                        PurePath(
                            path_to_annotations.replace(
                                iterated_list[counter - 1], splitted
                            )
                        )
                    )
                elif name_of_annotations_folder != "labels" and counter == 0:
                    path_to_annotations = str(
                        PurePath(
                            path_to_images.replace("images", name_of_annotations_folder)
                        )
                    )
                else:
                    path_to_annotations = str(
                        PurePath(path_to_images.replace("images", "labels"))
                    )

                path_to_images = str(PurePath("../../images/", splitted))

            if counter == 0:
                dataset = ImportYoloV5(
                    path=path_to_annotations,
                    path_to_images=path_to_images,
                    cat_names=yoloclasses,
                    img_ext=image_ext,
                )
                dataset.df["split"] = splitted
                counter += 1
            else:
                dataset2 = ImportYoloV5(
                    path=path_to_annotations,
                    path_to_images=path_to_images,
                    cat_names=yoloclasses,
                    img_ext=image_ext,
                )
                dataset2.df["split"] = splitted
                dataset.df = dataset.df.append(dataset2.df)
                counter += 1
    return dataset



import numpy as np

#########
# This file has variables and functions that are used by the rest of the package
#########

# These are the valid columns in the pylabel annotations table.
schema = [
    "img_folder",
    "img_filename",
    "img_path",
    "img_id",
    "img_width",
    "img_height",
    "img_depth",
    "ann_segmented",
    "ann_bbox_xmin",
    "ann_bbox_ymin",
    "ann_bbox_xmax",
    "ann_bbox_ymax",
    "ann_bbox_width",
    "ann_bbox_height",
    "ann_area",
    "ann_segmentation",
    "ann_iscrowd",
    "ann_pose",
    "ann_truncated",
    "ann_difficult",
    "cat_id",
    "cat_name",
    "cat_supercategory",
    "split",
    "annotated",
]
# schema = ['id','img_folder','img_filename','img_path','img_id','img_width','img_height','img_depth','ann_segmented','ann_bbox_xmin','ann_bbox_ymin','ann_bbox_xmax','ann_bbox_ymax','ann_bbox_width','ann_bbox_height','ann_area','ann_segmentation','ann_iscrowd','ann_pose','ann_truncated','ann_difficult','cat_id','cat_name','cat_supercategory','split']


def _ReindexCatIds(df, cat_id_index=0):
    """
    Reindex the values of the cat_id column so that that they start from an int (usually 0 or 1) and
    then increment the cat_ids to index + number of categories.
    It's useful if the cat_ids are not continuous, especially for dataset subsets,
    or combined multiple datasets. Some models like Yolo require starting from 0 and others
    like Detectron require starting from 1.
    """
    assert isinstance(cat_id_index, int), "cat_id_index must be an int."
    df["cat_id"] = df["cat_id"].astype("int")
    # Map cat_ids to the range [cat_id_index, cat_id_index + num_cats)
    unique_ids = np.sort(df["cat_id"].unique())
    ids_dict = dict((v, k) for k, v in enumerate(unique_ids, start=cat_id_index))
    df["cat_id"] = df["cat_id"].map(ids_dict)



import numpy as np
import pandas as pd
from sklearn.model_selection import GroupShuffleSplit as sklearnGroupShuffleSplit
from pylabel.shared import schema

class Split():
    def  __init__(self, dataset=None):
        self.dataset = dataset

    def UnSplit(self):
        """Unsplit the dataset by setting all values of the split column to null."""
        self.dataset.df["split"] = np.nan

    def GroupShuffleSplit(self, train_pct=.5, test_pct=.25, val_pct=.25, group_col = 'img_filename', random_state=None):
        """
        This function uses the GroupShuffleSplit command from sklearn. It can split into 3 groups (train,
        test, and val) by applying the command twice. 
        """
        df_main = self.dataset.df
        gss = sklearnGroupShuffleSplit(n_splits=1, train_size=train_pct)
        train_indexes, test_indexes = next(gss.split(X=df_main, y=df_main[group_col], groups=df_main.index.values))

        df_main.loc[train_indexes,'split'] = "train"
        df_main.loc[test_indexes,'split'] = "test"
        self.dataset.df = df_main

        if val_pct:
            df_train = df_main.loc[df_main['split'] == 'train']
            df_test = df_main.loc[df_main['split'] == 'test']
            df_test = df_test.reset_index()
            second_split_pct = float(test_pct/(test_pct+val_pct))
            gss2 = sklearnGroupShuffleSplit(n_splits=1, train_size=second_split_pct)
            test_indexes_2, val_indexes_2 = next(gss2.split(X=df_test, y=df_test[group_col], groups=df_test.index.values))
            df_test.loc[test_indexes_2,'split'] = "test"
            df_test.loc[val_indexes_2,'split'] = "val"
            self.dataset.df = df_train.append(df_test)
        self.dataset.df = self.dataset.df.reset_index(drop=True)
        self.dataset.df = self.dataset.df[schema]

    #Written with the help of https://stackoverflow.com/questions/56872664/complex-dataset-split-stratifiedgroupshufflesplit 
    def StratifiedGroupShuffleSplit(self, train_pct=.7, test_pct=.3, val_pct=.0, weight=0.01, 
        group_col = 'img_filename', cat_col = 'cat_name', batch_size=1):
        """
        This function will 'split" the dataframe by setting the split collumn equal to 
        train, test, or val. When a split dataset is exported the annotations will be split into
        seperate groups so that can be used used in model training, testing, and validation.
        """
        df_main = self.dataset.df
        df_main = df_main.reindex(np.random.permutation(df_main.index)) # shuffle dataset

        # create empty train, val and test datasets
        df_train = pd.DataFrame()
        df_val = pd.DataFrame()
        df_test = pd.DataFrame()

        subject_grouped_df_main = df_main.groupby([group_col], sort=False, as_index=False)
        category_grouped_df_main = df_main.groupby(cat_col).count()[[group_col]]/len(df_main)*100

        #Check inputs 
        assert (0 <= weight <= 1), "Weight must be between 0 and 1"
        total_splits = round((train_pct) + float(test_pct) + float(val_pct),1)
        assert (total_splits == 1), "Sum of train_pct, test_pct, and val_pct must equal 1."
        assert (batch_size >= 1 and batch_size <= subject_grouped_df_main.ngroups / 10 ), \
            "Batch must be greater than 1 and less than 1/10 count of groups"

        def calc_mse_loss(df):
            grouped_df = df.groupby(cat_col).count()[[group_col]]/len(df)*100
            df_temp = category_grouped_df_main.join(grouped_df, on = cat_col, how = 'left', lsuffix = '_main')
            df_temp.fillna(0, inplace=True)
            df_temp['diff'] = (df_temp['img_filename_main'] - df_temp[group_col])**2
            mse_loss = np.mean(df_temp['diff'])
            return mse_loss

        i = 0 #counter for all items in dataset
        b = 0 #counter for the batches
        batch_df = df_main[0:0]

        for _, group in subject_grouped_df_main:
            if (i < 3):
                if (i == 0):
                    df_train = df_train.append(pd.DataFrame(group), ignore_index=True)
                    i += 1
                    continue
                elif (i == 1):
                    df_val = df_val.append(pd.DataFrame(group), ignore_index=True)
                    i += 1
                    continue
                else:
                    df_test = df_test.append(pd.DataFrame(group), ignore_index=True)
                    i += 1
                    continue

            #Add groups to the 
            batch_df = batch_df.append(group)
            b += 1
            if b < batch_size and i < subject_grouped_df_main.ngroups-3:
                i += 1
                continue

            mse_loss_diff_train = calc_mse_loss(df_train) - calc_mse_loss(df_train.append(batch_df, ignore_index=True))
            mse_loss_diff_val = calc_mse_loss(df_val) - calc_mse_loss(df_val.append(batch_df, ignore_index=True))
            mse_loss_diff_test = calc_mse_loss(df_test) - calc_mse_loss(df_test.append(batch_df, ignore_index=True))

            total_records = len(df_train) + len(df_val) + len(df_test)

            len_diff_train = (train_pct - (len(df_train)/total_records))
            len_diff_val = (val_pct - (len(df_val)/total_records))
            len_diff_test = (test_pct - (len(df_test)/total_records)) 

            len_loss_diff_train = len_diff_train * abs(len_diff_train)
            len_loss_diff_val = len_diff_val * abs(len_diff_val)
            len_loss_diff_test = len_diff_test * abs(len_diff_test)

            loss_train = (weight * mse_loss_diff_train) + ((1-weight) * len_loss_diff_train)
            loss_val = (weight * mse_loss_diff_val) + ((1-weight) * len_loss_diff_val)
            loss_test = (weight * mse_loss_diff_test) + ((1-weight) * len_loss_diff_test)

            if (max(loss_train,loss_val,loss_test) == loss_train):
                df_train = df_train.append(batch_df, ignore_index=True)
            elif (max(loss_train,loss_val,loss_test) == loss_val):
                df_val = df_val.append(batch_df, ignore_index=True)
            else:
                df_test = df_test.append(batch_df, ignore_index=True)

            #print ("Group " + str(i) + ". loss_train: " + str(loss_train) + " | " + "loss_val: " + str(loss_val) + " | " + "loss_test: " + str(loss_test) + " | ")
            i += 1
            #Reset the batch
            b = 0
            batch_df = df_main[0:0]


        ######
        # Final prep tasks before returning the split dataframe

        #Sometimes the algo will put some rows in the val set even if the split percent was set to zero
        #In those cases move the rows from val to test 
        if round(val_pct,1) == round(0,1):
            df_test.append(df_val)
            df_val = df_val[0:0] #remove the values from 

        #Apply train, split, val labels to the split collumn 
        df_train['split'] = 'train'
        df_test['split'] = 'test'
        df_val['split'] = 'val'

        df = df_train.append(df_test).append(df_val)
        
        assert df.shape == df_main.shape, "Output shape does not match input shape. Data loss has occured."

        self.dataset.df = df
        self.dataset.df = self.dataset.df.reset_index(drop=True)
        self.dataset.df = self.dataset.df[schema]


"""PyLabel currently supports exporting annotations in COCO, YOLO, and VOC PASCAL formats."""

import json
from typing import List
import pandas as pd
import numpy as np
import xml.etree.ElementTree as ET
import xml.dom.minidom
import os
import yaml
import shutil
from pylabel.shared import _ReindexCatIds

from pathlib import PurePath, Path


class Export:
    def __init__(self, dataset=None):
        self.dataset = dataset

    def ExportToVoc(
        self,
        output_path=None,
        segmented_=False,
        path_=False,
        database_=False,
        folder_=False,
        occluded_=False,
    ):
        """Writes annotation files to disk in VOC XML format and returns path to files.

        By default, tags with empty values will not be included in the XML output.
        You can optionally choose to include them if they are required for your solution.

        Args:
            output_path (str):
                This is where the annotation files will be written.
                If not-specified then the path will be derived from the .path_to_annotations and
                .name properties of the dataset object.
            segmented_ (bool) :
                Defaults to False. Set to true to include this field in the XML schema of the output files.
            path_ (bool) :
                Defaults to False. Set to true to include this field in the XML schema of the output files.
            database_ (bool) :
                Defaults to False. Set to true to include this field in the XML schema of the output files.
            folder_ (bool) :
                Defaults to False. Set to true to include this field in the XML schema of the output files.
            occluded_ (bool) :
                Defaults to False. Set to true to include this field in the XML schema of the output files.

        Returns:
            A list with 1 or more paths (strings) to annotations files.

        Example:
            >>> dataset.export.ExportToVoc()
            ['data/voc_annotations/000000000322.xml', ...]
        """
        ds = self.dataset

        if output_path == None:
            output_path = ds.path_to_annotations
        else:
            output_path = output_path

        os.makedirs(output_path, exist_ok=True)

        output_file_paths = []

        def voc_xml_file_creation(
            data,
            file_name,
            output_file_path,
            segmented=True,
            path=True,
            database=True,
            folder=True,
            occluded=True,
        ):

            index = 0
            df_smaller = data[data["img_filename"] == file_name].reset_index()

            if len(df_smaller) == 1:
                # print('test')
                annotation_text_start = "<annotation>"

                flder_lkp = str(df_smaller.loc[index]["img_folder"])
                if folder == True and flder_lkp != "":
                    folder_text = "<folder>" + flder_lkp + "</folder>"
                else:
                    folder_text = ""

                filename_text = (
                    "<filename>"
                    + str(df_smaller.loc[index]["img_filename"])
                    + "</filename>"
                )

                pth_lkp = str(df_smaller.loc[index]["img_path"])
                if path == True and pth_lkp != "":
                    path_text = "<path>" + pth_lkp + "</path>"
                else:
                    path_text = ""

                sources_text = ""

                size_text_start = "<size>"
                width_text = (
                    "<width>" + str(df_smaller.loc[index]["img_width"]) + "</width>"
                )
                height_text = (
                    "<height>" + str(df_smaller.loc[index]["img_height"]) + "</height>"
                )
                depth_text = (
                    "<depth>" + str(df_smaller.loc[index]["img_depth"]) + "</depth>"
                )
                size_text_end = "</size>"

                seg_lkp = str(df_smaller.loc[index]["ann_segmented"])
                if segmented == True and seg_lkp != "":
                    segmented_text = (
                        "<segmented>"
                        + str(df_smaller.loc[index]["ann_segmented"])
                        + "</segmented>"
                    )
                else:
                    segmented_text = ""

                object_text_start = "<object>"

                name_text = (
                    "<name>" + str(df_smaller.loc[index]["cat_name"]) + "</name>"
                )
                pose_text = (
                    "<pose>" + str(df_smaller.loc[index]["ann_pose"]) + "</pose>"
                )
                truncated_text = (
                    "<truncated>"
                    + str(df_smaller.loc[index]["ann_truncated"])
                    + "</truncated>"
                )
                difficult_text = (
                    "<difficult>"
                    + str(df_smaller.loc[index]["ann_difficult"])
                    + "</difficult>"
                )

                occluded_text = ""

                bound_box_text_start = "<bndbox>"

                xmin_text = (
                    "<xmin>"
                    + str(df_smaller.loc[index]["ann_bbox_xmin"].astype("int"))
                    + "</xmin>"
                )
                xmax_text = (
                    "<xmax>"
                    + str(df_smaller.loc[index]["ann_bbox_xmax"].astype("int"))
                    + "</xmax>"
                )
                ymin_text = (
                    "<ymin>"
                    + str(df_smaller.loc[index]["ann_bbox_ymin"].astype("int"))
                    + "</ymin>"
                )
                ymax_text = (
                    "<ymax>"
                    + str(df_smaller.loc[index]["ann_bbox_ymax"].astype("int"))
                    + "</ymax>"
                )

                bound_box_text_end = "</bndbox>"
                object_text_end = "</object>"
                annotation_text_end = "</annotation>"

                xmlstring = (
                    annotation_text_start
                    + folder_text
                    + filename_text
                    + path_text
                    + sources_text
                    + size_text_start
                    + width_text
                    + height_text
                    + depth_text
                    + size_text_end
                    + segmented_text
                    + object_text_start
                    + name_text
                    + pose_text
                    + truncated_text
                    + difficult_text
                    + occluded_text
                    + bound_box_text_start
                    + xmin_text
                    + xmax_text
                    + ymin_text
                    + ymax_text
                    + bound_box_text_end
                    + object_text_end
                    + annotation_text_end
                )
                dom = xml.dom.minidom.parseString(xmlstring)
                pretty_xml_as_string = dom.toprettyxml()

                with open(output_file_path, "w") as f:
                    f.write(pretty_xml_as_string)

                return output_file_path

            else:

                # print('test')
                annotation_text_start = "<annotation>"

                flder_lkp = str(df_smaller.loc[index]["img_folder"])
                if folder == True and flder_lkp != "":
                    folder_text = "<folder>" + flder_lkp + "</folder>"
                else:
                    folder_text = ""

                filename_text = (
                    "<filename>"
                    + str(df_smaller.loc[index]["img_filename"])
                    + "</filename>"
                )

                pth_lkp = str(df_smaller.loc[index]["img_path"])
                if path == True and pth_lkp != "":
                    path_text = "<path>" + pth_lkp + "</path>"
                else:
                    path_text = ""

                # db_lkp = str(df_smaller.loc[index]['Databases'])
                # if database == True and db_lkp != '':
                #    sources_text = '<source>'+'<database>'+ db_lkp +'</database>'+'</source>'
                # else:
                sources_text = ""

                size_text_start = "<size>"
                width_text = (
                    "<width>" + str(df_smaller.loc[index]["img_width"]) + "</width>"
                )
                height_text = (
                    "<height>" + str(df_smaller.loc[index]["img_height"]) + "</height>"
                )
                depth_text = (
                    "<depth>" + str(df_smaller.loc[index]["img_depth"]) + "</depth>"
                )
                size_text_end = "</size>"

                seg_lkp = str(df_smaller.loc[index]["ann_segmented"])
                if segmented == True and seg_lkp != "":
                    segmented_text = (
                        "<segmented>"
                        + str(df_smaller.loc[index]["ann_segmented"])
                        + "</segmented>"
                    )
                else:
                    segmented_text = ""

                xmlstring = (
                    annotation_text_start
                    + folder_text
                    + filename_text
                    + path_text
                    + sources_text
                    + size_text_start
                    + width_text
                    + height_text
                    + depth_text
                    + size_text_end
                    + segmented_text
                )

                for obj in range(len(df_smaller)):
                    object_text_start = "<object>"

                    name_text = (
                        "<name>" + str(df_smaller.loc[index]["cat_name"]) + "</name>"
                    )
                    pose_text = (
                        "<pose>" + str(df_smaller.loc[index]["ann_pose"]) + "</pose>"
                    )
                    truncated_text = (
                        "<truncated>"
                        + str(df_smaller.loc[index]["ann_truncated"])
                        + "</truncated>"
                    )
                    difficult_text = (
                        "<difficult>"
                        + str(df_smaller.loc[index]["ann_difficult"])
                        + "</difficult>"
                    )

                    # occ_lkp = str(df_smaller.loc[index]['Object Occluded'])
                    # if occluded==True and occ_lkp != '':
                    #    occluded_text = '<occluded>'+occ_lkp+'</occluded>'
                    # else:
                    occluded_text = ""

                    bound_box_text_start = "<bndbox>"

                    xmin_text = (
                        "<xmin>"
                        + str(df_smaller.loc[index]["ann_bbox_xmin"].astype("int"))
                        + "</xmin>"
                    )
                    xmax_text = (
                        "<xmax>"
                        + str(df_smaller.loc[index]["ann_bbox_xmax"].astype("int"))
                        + "</xmax>"
                    )
                    ymin_text = (
                        "<ymin>"
                        + str(df_smaller.loc[index]["ann_bbox_ymin"].astype("int"))
                        + "</ymin>"
                    )
                    ymax_text = (
                        "<ymax>"
                        + str(df_smaller.loc[index]["ann_bbox_ymax"].astype("int"))
                        + "</ymax>"
                    )

                    bound_box_text_end = "</bndbox>"
                    object_text_end = "</object>"
                    annotation_text_end = "</annotation>"
                    index = index + 1

                    xmlstring = (
                        xmlstring
                        + object_text_start
                        + name_text
                        + pose_text
                        + truncated_text
                        + difficult_text
                        + occluded_text
                        + bound_box_text_start
                        + xmin_text
                        + xmax_text
                        + ymin_text
                        + ymax_text
                        + bound_box_text_end
                        + object_text_end
                    )

                xmlstring = xmlstring + annotation_text_end
                dom = xml.dom.minidom.parseString(xmlstring)
                pretty_xml_as_string = dom.toprettyxml()

                with open(output_file_path, "w") as f:
                    f.write(pretty_xml_as_string)

                return output_file_path

        # Loop through all images in the dataframe and call voc_xml_file_creation for each one
        for file_title in list(set(self.dataset.df.img_filename)):

            file_name = Path(file_title)
            file_name = str(file_name.with_suffix(".xml"))
            file_path = str(Path(output_path, file_name))
            voc_file_path = voc_xml_file_creation(
                ds.df,
                file_title,
                segmented=segmented_,
                path=path_,
                database=database_,
                folder=folder_,
                occluded=occluded_,
                output_file_path=file_path,
            )
            output_file_paths.append(voc_file_path)

        return output_file_paths

    def ExportToYoloV5(
        self,
        output_path="training/labels",
        yaml_file="dataset.yaml",
        copy_images=False,
        use_splits=False,
        cat_id_index=None,
    ):
        """Writes annotation files to disk in YOLOv5 format and returns the paths to files.

        Args:

            output_path (str):
                This is where the annotation files will be written.
                If not-specified then the path will be derived from the .path_to_annotations and
                .name properties of the dataset object. If you are exporting images to train a model, the recommended path
                to use is 'training/labels'.
            yaml_file (str):
                If a file name (string) is provided, a YOLOv5 YAML file will be created with entries for the files
                and classes in this dataset. It will be created in the parent of the output_path directory.
                The recommended name for the YAML file is 'dataset.yaml'.
            copy_images (boolean):
                If True, then the annotated images will be copied to a directory next to the labels directory into
                a directory named 'images'. This will prepare your labels and images to be used as inputs to
                train a YOLOv5 model.
            use_splits (boolean):
                If True, then the images and annotations will be moved into directories based on the values in the split column.
                For example, if a row has the value split = "train" then the annotations for that row will be moved to directory
                /train. If a YAML file is specificied then the YAML file will use the splits to specify the folders user for the
                train, val, and test datasets.
            cat_id_index (int):
                Reindex the cat_id values so that that they start from an int (usually 0 or 1) and
                then increment the cat_ids to index + number of categories continuously.
                It's useful if the cat_ids are not continuous in the original dataset.
                Yolo requires the set of annotations to start at 0 when training a model.

        Returns:
            A list with 1 or more paths (strings) to annotations files. If a YAML file is created
            then the first item in the list will be the path to the YAML file.

        Examples:
            >>> dataset.export.ExportToYoloV5(output_path='training/labels',
            >>>     yaml_file='dataset.yaml', cat_id_index=0)
            ['training/dataset.yaml', 'training/labels/frame_0002.txt', ...]

        """
        ds = self.dataset

        # Inspired by https://github.com/aws-samples/groundtruth-object-detection/blob/master/create_annot.py
        yolo_dataset = ds.df.copy(deep=True)
        # Convert nan values in the split column from nan to '' because those are easier to with with when building paths
        yolo_dataset.split = yolo_dataset.split.fillna("")

        # Create all of the paths that will be used to manage the files in this dataset
        path_dict = {}

        # The output path is the main path that will be used to create the other relative paths
        path = PurePath(output_path)
        path_dict["label_path"] = output_path
        # The /images directory should be next to the /labels directory
        path_dict["image_path"] = str(PurePath(path.parent, "images"))
        # The root directory is in parent of the /labels and /images directories
        path_dict["root_path"] = str(PurePath(path.parent))
        # The YAML file should be in root directory
        path_dict["yaml_path"] = str(PurePath(path_dict["root_path"], yaml_file))
        # The root directory will usually be next to the yolov5 directory.
        # Specify the relative path
        path_dict["root_path_from_yolo_dir"] = str(PurePath("../"))
        # If these default values to not match the users environment then they can manually edit the YAML file

        if copy_images:
            # Create the folder that the images will be copied to
            Path(path_dict["image_path"]).mkdir(parents=True, exist_ok=True)

        # Drop rows that are not annotated
        # Note, having zero annotates can still be considered annotated
        # in cases when are no objects in the image thats should be indentified
        yolo_dataset = yolo_dataset.loc[yolo_dataset["annotated"] == 1]
        yolo_dataset["cat_id"] = (
            yolo_dataset["cat_id"].astype("float").astype(pd.Int32Dtype())
        )

        if cat_id_index != None:
            assert isinstance(cat_id_index, int), "cat_id_index must be an int."
            _ReindexCatIds(yolo_dataset, cat_id_index)

        yolo_dataset["center_x_scaled"] = (
            yolo_dataset["ann_bbox_xmin"] + (yolo_dataset["ann_bbox_width"] * 0.5)
        ) / yolo_dataset["img_width"]
        yolo_dataset["center_y_scaled"] = (
            yolo_dataset["ann_bbox_ymin"] + (yolo_dataset["ann_bbox_height"] * 0.5)
        ) / yolo_dataset["img_height"]
        yolo_dataset["width_scaled"] = (
            yolo_dataset["ann_bbox_width"] / yolo_dataset["img_width"]
        )
        yolo_dataset["height_scaled"] = (
            yolo_dataset["ann_bbox_height"] / yolo_dataset["img_height"]
        )
        yolo_dataset[
            [
                "cat_id",
                "center_x_scaled",
                "center_y_scaled",
                "width_scaled",
                "height_scaled",
            ]
        ]

        # Create folders to store annotations
        if output_path == None:
            dest_folder = PurePath(
                ds.path_to_annotations, yolo_dataset.iloc[0].img_folder
            )
        else:
            dest_folder = output_path

        os.makedirs(dest_folder, exist_ok=True)

        unique_images = yolo_dataset["img_filename"].unique()
        output_file_paths = []

        for img_filename in unique_images:
            df_single_img_annots = yolo_dataset.loc[
                yolo_dataset.img_filename == img_filename
            ]
            annot_txt_file = img_filename.split(".")[0] + ".txt"
            # Use the value of the split collumn to create a directory
            # The values should be train, val, test or ''
            if use_splits:
                split_dir = df_single_img_annots.iloc[0].split
            else:
                split_dir = ""
            destination = str(PurePath(dest_folder, split_dir, annot_txt_file))
            Path(
                dest_folder,
                split_dir,
            ).mkdir(parents=True, exist_ok=True)

            df_single_img_annots.to_csv(
                destination,
                index=False,
                header=False,
                sep=" ",
                float_format="%.4f",
                columns=[
                    "cat_id",
                    "center_x_scaled",
                    "center_y_scaled",
                    "width_scaled",
                    "height_scaled",
                ],
            )
            output_file_paths.append(destination)

            if copy_images:
                source_image_path = str(
                    Path(
                        ds.path_to_annotations,
                        df_single_img_annots.iloc[0].img_folder,
                        df_single_img_annots.iloc[0].img_filename,
                    )
                )

                current_file = Path(source_image_path)
                assert (
                    current_file.is_file
                ), f"File does not exist: {source_image_path}. Check img_folder column values."
                Path(path_dict["image_path"], split_dir).mkdir(
                    parents=True, exist_ok=True
                )
                shutil.copy(
                    str(source_image_path),
                    str(PurePath(path_dict["image_path"], split_dir, img_filename)),
                )

        # Create YAML file
        if yaml_file:
            # Make a set with all of the different values of the split column
            splits = set(yolo_dataset.split)
            # Build a dict with all of the values that will go into the YAML file
            dict_file = {}
            dict_file["path"] = path_dict["root_path_from_yolo_dir"]

            # If train is one of the splits, append train to path
            if use_splits and "train" in splits:
                dict_file["train"] = str(PurePath(path_dict["image_path"], "train"))
            else:
                dict_file["train"] = path_dict["image_path"]

            # If val is one of the splits, append val to path
            if use_splits and "val" in splits:
                dict_file["val"] = str(PurePath(path_dict["image_path"], "val"))
            else:
                # If there is no val split, use the train split as the val split
                dict_file["val"] = dict_file["train"]

            # If test is one of the splits, make a test param and add test to the path
            if use_splits and "test" in splits:
                dict_file["test"] = str(PurePath(path_dict["image_path"], "test"))

            dict_file["nc"] = ds.analyze.num_classes
            dict_file["names"] = ds.analyze.classes

            # Save the yamlfile
            with open(path_dict["yaml_path"], "w") as file:
                documents = yaml.dump(dict_file, file)
                output_file_paths = [path_dict["yaml_path"]] + output_file_paths

        return output_file_paths

    def ExportToCoco(self, output_path=None, cat_id_index=None):
        """
        Writes COCO annotation files to disk (in JSON format) and returns the path to files.

        Args:
            output_path (str):
                This is where the annotation files will be written. If not-specified then the path will be derived from the path_to_annotations and
                name properties of the dataset object.
            cat_id_index (int):
                Reindex the cat_id values so that that they start from an int (usually 0 or 1) and
                then increment the cat_ids to index + number of categories continuously.
                It's useful if the cat_ids are not continuous in the original dataset.
                Some models like Yolo require starting from 0 and others like Detectron require starting from 1.

        Returns:
            A list with 1 or more paths (strings) to annotations files.

        Example:
            >>> dataset.exporter.ExportToCoco()
            ['data/labels/dataset.json']

        """
        # Copy the dataframe in the dataset so the original dataset doesn't change when you apply the export tranformations
        df = self.dataset.df.copy(deep=True)
        # Tweak dataset for standardized output
        df["cat_id"] = df["cat_id"].astype("int")
        df["ann_iscrowd"] = df["ann_iscrowd"].fillna(0)

        if cat_id_index != None:
            assert isinstance(cat_id_index, int), "cat_id_index must be an int."
            _ReindexCatIds(df, cat_id_index)

        df_outputI = []
        df_outputA = []
        df_outputC = []
        list_i = []
        list_c = []
        json_list = []

        for i in range(0, df.shape[0]):
            images = [
                {
                    "id": df["img_id"][i],
                    "folder": df["img_folder"][i],
                    "file_name": df["img_filename"][i],
                    "path": df["img_path"][i],
                    "width": df["img_width"][i],
                    "height": df["img_height"][i],
                    "depth": df["img_depth"][i],
                }
            ]

            annotations = [
                {
                    "image_id": df["img_id"][i],
                    "id": df.index[i],
                    "segmented": df["ann_segmented"][i],
                    "bbox": [
                        df["ann_bbox_xmin"][i],
                        df["ann_bbox_ymin"][i],
                        df["ann_bbox_width"][i],
                        df["ann_bbox_height"][i],
                    ],
                    "area": df["ann_area"][i],
                    "segmentation": df["ann_segmentation"][i],
                    "iscrowd": df["ann_iscrowd"][i],
                    "pose": df["ann_pose"][i],
                    "truncated": df["ann_truncated"][i],
                    "category_id": df["cat_id"][i],
                    "difficult": df["ann_difficult"][i],
                }
            ]

            categories = [
                {
                    "id": int(df["cat_id"][i]),
                    "name": df["cat_name"][i],
                    "supercategory": df["cat_supercategory"][i],
                }
            ]

            if list_c:
                if categories[0]["id"] in list_c or np.isnan(categories[0]["id"]):
                    pass
                else:
                    df_outputC.append(pd.DataFrame([categories]))
            elif ~np.isnan(categories[0]["id"]):
                df_outputC.append(pd.DataFrame([categories]))
            else:
                pass
            list_c.append(categories[0]["id"])

            if list_i:
                if images[0]["id"] in list_i or np.isnan(images[0]["id"]):
                    pass
                else:
                    df_outputI.append(pd.DataFrame([images]))
            elif ~np.isnan(images[0]["id"]):
                df_outputI.append(pd.DataFrame([images]))
            else:
                pass
            list_i.append(images[0]["id"])

            df_outputA.append(pd.DataFrame([annotations]))

        mergedI = pd.concat(df_outputI, ignore_index=True)
        mergedA = pd.concat(df_outputA, ignore_index=True)
        mergedC = pd.concat(df_outputC, ignore_index=True)

        resultI = mergedI[0].to_json(orient="split", default_handler=str)
        resultA = mergedA[0].to_json(orient="split", default_handler=str)
        resultC = mergedC[0].to_json(orient="split", default_handler=str)

        parsedI = json.loads(resultI)
        del parsedI["index"]
        del parsedI["name"]
        parsedI["images"] = parsedI["data"]
        del parsedI["data"]

        parsedA = json.loads(resultA)
        del parsedA["index"]
        del parsedA["name"]
        parsedA["annotations"] = parsedA["data"]
        del parsedA["data"]

        parsedC = json.loads(resultC)
        del parsedC["index"]
        del parsedC["name"]
        parsedC["categories"] = parsedC["data"]
        del parsedC["data"]

        parsedI.update(parsedA)
        parsedI.update(parsedC)
        json_output = parsedI

        if output_path == None:
            output_path = Path(
                self.dataset.path_to_annotations, (self.dataset.name + ".json")
            )

        with open(output_path, "w") as outfile:
            json.dump(obj=json_output, fp=outfile, indent=4)
        return [str(output_path)]



import bbox_visualizer as bbv
import cv2
from PIL import Image
from pathlib import Path

class Visualize:
    def  __init__(self, dataset=None):
        self.dataset = dataset

    """Functions to visualize inspect images and annotations."""
    def ShowBoundingBoxes(self, img_id:int=0, img_filename:str="") -> Image:
        """Enter a filename or index number and return the image with the bounding boxes drawn."""

        ds = self.dataset

        #Handle cases where user enters image name in default field
        if type(img_id) == str:
            img_filename = img_id

        if img_filename == "": 
            df_single_img_annots = ds.df.loc[ds.df.img_id == img_id]
        else:
            df_single_img_annots = ds.df.loc[ds.df.img_filename == img_filename]

        full_image_path = str(Path(ds.path_to_annotations, df_single_img_annots.iloc[0].img_folder, df_single_img_annots.iloc[0].img_filename))
        img = cv2.imread(str(full_image_path))
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        labels = []
        bboxes = []

        for index, row in df_single_img_annots.iterrows():
            labels.append(row['cat_name'])
            bboxes.append([int(row['ann_bbox_xmin']),int(row['ann_bbox_ymin']),int(row['ann_bbox_xmax']),int(row['ann_bbox_ymax'])])

        img_with_boxes = bbv.draw_multiple_rectangles(img, bboxes)
        img_with_boxes = bbv.add_multiple_labels(img_with_boxes, labels, bboxes)

        rendered_img = Image.fromarray(img_with_boxes)
        #rendered_img.save("bbox-visualizer/jpeg.jpg")
        return rendered_img




"""The dataset is the primary object that you will interactive with when using PyLabel.
All other modules are sub-modules of the dataset object. 
"""

from pylabel.analyze import Analyze
from pylabel.exporter import Export
from pylabel.visualize import Visualize
from pylabel.labeler import Labeler
from pylabel.splitter import Split
from pylabel.shared import _ReindexCatIds

import numpy as np


class Dataset:
    def __init__(self, df):
        self.df = df
        """Pandas Dataframe: The dataframe where annotations are stored. This dataframe can be read directly
        to query the contents of the dataset. You can also edit this dataframe to filter records or edit the 
        annotations directly. 

        Example: 
            >>> dataset.df
        """
        self.name = "dataset"
        """string: Default is 'dataset'. A friendly name for your dataset that is used as part of the filename(s)
        when exporting annotation files. 
        """
        self.path_to_annotations = ""
        """string: Default is ''. The path to the annotation files associated with the dataset. When importing, 
        this will be path to the directory where the annotations are stored.  By default, annotations will be exported
        to the same directory. Changing this value will change where the annotations are exported to.  
        """
        self.export = Export(dataset=self)
        """See pylabel.export module."""
        self.visualize = Visualize(dataset=self)
        """See pylabel.visualize module."""
        self.analyze = Analyze(dataset=self)
        """See pylabel.analyze module."""
        self.labeler = Labeler(self)
        """See pylabel.labeler module."""
        self.splitter = Split(dataset=self)
        """See pylabel.splitter module."""

    def ReindexCatIds(self, cat_id_index=0):
        """
        Reindex the values of the cat_id column so that that they start from an int (usually 0 or 1) and
        then increment the cat_ids to index + number of categories.
        It's useful if the cat_ids are not continuous, especially for dataset subsets,
        or combined multiple datasets. Some models like Yolo require starting from 0 and others
        like Detectron require starting from 1.

        Args:
            cat_id_index (int): Defaults to 0.
                The cat ids will increment sequentially the cat_index value. For example if there are 10
                classes then the cat_ids will be a range from 0-9.

        Example:
            >>> dataset.analyze.class_ids
                [1,2,4,5,6,7,8,9,11,12]
            >>> dataset.ReindexCatIds(cat_id_index) = 0
            >>> dataset.analyze.class_ids
                [0,1,2,3,4,5,6,7,8,9]
        """
        assert isinstance(cat_id_index, int), "cat_id_index must be an int."
        _ReindexCatIds(self.df, cat_id_index)



# Test
from setuptools import setup


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="pylabel",
    packages=["pylabel"],
    version="0.1.31",
    description="Transform, analyze, and visualize computer vision annotations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pylabel-project/pylabel",
    author="PyLabel Project",
    license="MIT",
    install_requires=[
        "pandas",
        "bbox_visualizer",
        "matplotlib",
        "opencv-python",
        "scikit-learn",
        "jupyter_bbox_widget",
        "pyyaml",
    ],
)



"""The analyze module analyzes the contents of the dataset and provides summary statistics 
such as the count of images and classes. """

import pandas as pd
import numpy as np


class Analyze:
    def __init__(self, dataset=None):
        self.dataset = dataset
        # self.split_counts = ds.df["split"].value_counts(dropna=False)
        # self.split_pct = ds.df["split"].value_counts(normalize=True, dropna=False)

    @property
    def classes(self):
        """Returns list of all cat names in the dataset sorted by cat_id value.

        Returns:
            List

        Example:
            >>> dataset.analyze.classes
            ["Squirrel", "Nut"]
        """
        filtered_df = self.dataset.df[self.dataset.df["cat_id"].notnull()]
        categories = dict(zip(filtered_df.cat_name, filtered_df.cat_id.astype("int")))
        categories = sorted(categories.items(), key=lambda x: x[1])
        return [c[0] for c in categories if str(c[0]).strip() != ""]

    @property
    def class_ids(self):
        """Returns a sorted list of all cat ids in the dataset.

        Returns:
            List

        Example:
            >>> dataset.analyze.class_ids
            [0,1]
        """
        filtered_df = self.dataset.df[self.dataset.df["cat_id"].notnull()]
        cat_ids = list(filtered_df.cat_id.astype("int").unique())
        cat_ids.sort()
        return cat_ids

    @property
    def class_counts(self):
        """Counts the numbers of instances of each class in the dataset. Uses the Pandas value_counts
        method and returns a Pandas Series.

        Returns:
            Pandas Series
        Example:
            >>> dataset.analyze.class_counts
            squirrel  50
            nut       100
        """
        return self.dataset.df["cat_name"].value_counts(dropna=False)

    @property
    def num_classes(self):
        """Counts the unique number of classes in the dataset.

        Returns:
            Int
        Example:
            >>> dataset.analyze.num_classes
            2
        """
        cat_names = list(self.dataset.df.cat_name.unique())
        return len([i for i in cat_names if str(i).strip() != ""])

    @property
    def num_images(self):
        """Counts the number of images in the dataset.

        Returns:
            Int
        Example:
            >>> dataset.analyze.num_images
            100
        """
        return self.dataset.df["img_filename"].nunique()

    @property
    def class_name_id_map(self):
        """Returns a dict where the class name is the key and class id is the value.

        Returns:
            Dict

        Example:
            >>> dataset.analyze.class_name_id_map
            {('Squirrel', 0),('Nut', 1)}
        """
        return dict(zip(self.dataset.df.cat_name, self.dataset.df.cat_id))

    def ShowClassSplits(self, normalize=True):
        """Show the distribution of classes across train, val, and
        test splits of the dataset.

        Args:
            normalize (bool): Defaults to True.
                If True, then will return the relative frequencies of the classes between 0 and 1.
                If False, then will return the absolute counts of each class.

        Returns:
            Pandas Dataframe

        Examples:
            >>> dataset.analyze.ShowClassSplits(normalize=True)
            cat_name  all  train  test  dev
            squirrel  .66  .64    .65   .63
            nut       .34  .34    .35   .37

            >>> dataset.analyze.ShowClassSplits(normalize=False)
            cat_name  all  train  test  dev
            squirrel  66   64     65    63
            nut       34   34     35    37

        """
        ds = self.dataset

        def move_column_inplace(df, col, pos):
            """
            Assists to rearrange columns to a desired order.
            """
            col = df.pop(col)
            df.insert(pos, col.name, col)

        df_value_counts = pd.DataFrame(
            ds.df["cat_name"].value_counts(normalize=normalize), columns=["cat_name"]
        )

        df_value_counts.index.name = "cat_name"
        df_value_counts.columns = ["all"]

        split_df = ds.df.groupby("split")

        if split_df.ngroups == 1:
            return df_value_counts

        for name, group in split_df:
            group_df = pd.DataFrame(group)
            df_split_value_counts = pd.DataFrame(
                group_df["cat_name"].value_counts(normalize=normalize),
                columns=["cat_name"],
            )
            df_split_value_counts.index.name = "cat_name"
            df_split_value_counts.columns = [name]
            df_value_counts = pd.merge(
                df_value_counts, df_split_value_counts, how="left", on=["cat_name"]
            )

        # Move 'train' to the left of the table since that is the usual convention.
        if "train" in df_value_counts.columns:
            move_column_inplace(df_value_counts, "train", 1)

        return df_value_counts



import base64
from types import new_class
import ipywidgets as widgets
from pathlib import PurePath
from jupyter_bbox_widget import BBoxWidget
from ipywidgets import Layout
import numpy as np


class Labeler:
    def __init__(self, dataset=None):
        self.dataset = dataset

    def StartPyLaber(self, new_classes=None, image=None, yolo_model=None):
        """Display the bbox widget loaded with images and annotations from this dataset."""

        if "google.colab" in str(get_ipython()):
            from google.colab import output

            output.enable_custom_widget_manager()

        dataset = self.dataset
        widget_output = None

        files = dataset.df.img_filename.unique()
        files = files.tolist()

        global file_index

        if image == None:
            file_index = 0
            image = files[0]
        else:
            file_index = files.index(image)

        def GetBBOXs(image):
            # Make a dataframe with the annotations for a single image
            img_df = dataset.df.loc[dataset.df["img_filename"] == image]
            img_df_subset = img_df[
                [
                    "cat_name",
                    "ann_bbox_height",
                    "ann_bbox_width",
                    "ann_bbox_xmin",
                    "ann_bbox_ymin",
                ]
            ]
            # Rename the columns to match the format used by jupyter_bbox_widget
            img_df_subset.columns = ["label", "height", "width", "x", "y"]

            # Drop rows that have NaN, invalid bounding boxes
            img_df_subset = img_df_subset.dropna()
            bboxes_dict = img_df_subset.to_dict(orient="records")
            return bboxes_dict

        def GetImageLabel(image):
            """Returns the imagename or imagename + (not annotated)"""
            # Make a dataframe with the annotations for a single image
            img_df = dataset.df.loc[dataset.df["img_filename"] == image]

            if img_df.iloc[0].annotated == 1:
                return image
            else:
                return f"{image} (not annotated)"
            return

        bboxes_dict = GetBBOXs(image)

        img_folder = dataset.df.loc[dataset.df["img_filename"] == image].iloc[0][
            "img_folder"
        ]

        file_paths = [
            str(PurePath(dataset.path_to_annotations, img_folder, file))
            for file in files
        ]

        def encode_image(filepath):
            with open(filepath, "rb") as f:
                image_bytes = f.read()
            encoded = str(base64.b64encode(image_bytes), "utf-8")
            return "data:image/jpg;base64," + encoded

        def UpdateCategoryList(cat_dict, new_categories):
            from math import isnan

            # Remove invalid entries
            cat_dict.pop("", None)
            # cat_dict = {k: v for k, v in cat_dict.items() if not isnan(v)}

            for cat in new_categories:
                if len(cat_dict) == 0:
                    cat_dict[cat] = "0"
                elif cat in list(cat_dict.keys()):
                    continue
                else:
                    # Create a new cat id that 1+ the highest cat id value
                    new_cat_id = max([int(v) for v in cat_dict.values()]) + 1
                    cat_dict[cat] = str(new_cat_id)

            return cat_dict

        def on_submit(b):
            # save annotations for current image
            import pandas as pd

            global widget_output
            global file_index

            img_filename = files[file_index]

            widget_output = pd.DataFrame.from_dict(w_bbox.bboxes)

            # Check if there are any bounding boxes for the current image:
            if not widget_output.empty:
                # If there are bounding boxes then add them to the dataset
                widget_output = widget_output.rename(
                    columns={
                        "label": "cat_name",
                        "height": "ann_bbox_height",
                        "width": "ann_bbox_width",
                        "x": "ann_bbox_xmin",
                        "y": "ann_bbox_ymin",
                    }
                )

                widget_output["ann_area"] = (
                    widget_output["ann_bbox_height"] * widget_output["ann_bbox_width"]
                )
                widget_output["cat_name"] = widget_output["cat_name"].astype("string")
                categories = dict(zip(dataset.df.cat_name, dataset.df.cat_id))
                categories = UpdateCategoryList(
                    categories, list(widget_output.cat_name)
                )
                widget_output["cat_id"] = widget_output["cat_name"].map(categories)

            else:
                # Build a entry in the table with empty bounding box columns
                widget_output = pd.DataFrame(
                    [],
                    columns=[
                        "cat_name",
                        "cat_id",
                        "ann_bbox_height",
                        "ann_bbox_width",
                        "ann_bbox_xmin",
                        "ann_bbox_ymin",
                        "ann_area",
                    ],
                )

                widget_output.loc[0] = [
                    "",
                    np.NaN,
                    np.NaN,
                    np.NaN,
                    np.NaN,
                    np.NaN,
                    np.NaN,
                ]

            widget_output["img_filename"] = str(img_filename)
            widget_output["img_filename"] = widget_output["img_filename"].astype(
                "string"
            )
            widget_output.index.name = "id"

            img_df = dataset.df.loc[dataset.df["img_filename"] == img_filename]

            # Get the metadata associated with this image from the dataset
            metadata = img_df.iloc[0].to_frame().T

            metadata["img_filename"] = metadata["img_filename"].astype("string")

            # Drop the fields that are in the widget_output dataframe
            metadata.drop(
                [
                    "cat_name",
                    "cat_id",
                    "ann_area",
                    "ann_bbox_height",
                    "ann_bbox_width",
                    "ann_bbox_xmin",
                    "ann_bbox_ymin",
                ],
                axis=1,
                inplace=True,
            )

            widget_output = widget_output.merge(
                metadata, left_on="img_filename", right_on="img_filename"
            )
            widget_output = widget_output[dataset.df.columns]

            # Set annotated = 1, which means the annotates have been reviewed and accepted
            widget_output["annotated"] = 1

            # Now we have a dataframe with output of the bbox widget
            # Drop the current annotations for the image and add the the new ones
            dataset.df.drop(
                dataset.df[dataset.df["img_filename"] == img_filename].index,
                inplace=True,
            )
            dataset.df.reset_index(drop=True, inplace=True)

            dataset.df = dataset.df.append(widget_output).reset_index(drop=True)

            # move on to the next file
            on_next(b)

        def on_next(b):
            global file_index

            file_index += 1
            # open new image in the widget
            image_file = file_paths[file_index]
            w_bbox.image = encode_image(image_file)
            w_bbox.bboxes = GetBBOXs(files[file_index])
            progress_label.value = f"{file_index+1} / {len(files)}"
            current_img_name_label.value = GetImageLabel(files[file_index])

        def on_back(b):
            global file_index
            file_index -= 1
            # open new image in the widget
            image_file = file_paths[file_index]
            w_bbox.image = encode_image(image_file)
            w_bbox.bboxes = GetBBOXs(files[file_index])
            progress_label.value = f"{file_index+1} / {len(files)}"
            current_img_name_label.value = GetImageLabel(files[file_index])

        def on_add_class(b):
            if new_class_text.value.strip() != "":
                class_list = list(w_bbox.classes)
                class_list.append(new_class_text.value)
                w_bbox.classes = list(set(class_list))
                new_class_text.value = ""

        def on_predict(b):
            global file_index
            image_file = file_paths[file_index]

            result = yolo_model(image_file)
            result = result.pandas().xyxy[0]
            result["width"] = result.xmax - result.xmin
            result["height"] = result.ymax - result.ymin
            result.drop(["class", "confidence", "xmax", "ymax"], axis=1, inplace=True)
            result.columns = ["x", "y", "label", "width", "height"]
            result = result[["label", "height", "width", "x", "y"]]
            bboxes_dict = result.to_dict(orient="records")
            w_bbox.bboxes = bboxes_dict
            # Add the predicted classes to the widget so they can be selected by the user
            w_bbox.classes = list(set(w_bbox.classes + list(result["label"])))

        if new_classes:
            classes = dataset.analyze.classes + new_classes
        else:
            classes = dataset.analyze.classes

        # remove empty labels and duplicate labels
        classes = list(
            set([str(c).strip() for c in classes if len(str(c).strip()) > 0])
        )

        # Load BBoxWidget for first load on page
        w_bbox = BBoxWidget(
            image=encode_image(file_paths[file_index]),
            classes=classes,
            bboxes=bboxes_dict,
            hide_buttons=True,
        )

        progress_txt = f"{file_index+1} / {len(files)}"

        left_arrow_btn = widgets.Button(
            icon="fa-arrow-left", layout=Layout(width="35px")
        )
        progress_label = widgets.Label(value=progress_txt)
        current_img_name_label = widgets.Label(value=GetImageLabel(image))

        right_arrow_btn = widgets.Button(
            icon="fa-arrow-right", layout=Layout(width="35px")
        )
        save_btn = widgets.Button(
            icon="fa-check", description="Save", layout=Layout(width="70px")
        )
        predict_btn = widgets.Button(
            icon="fa-eye", description="Predict", layout=Layout(width="100px")
        )
        train_btn = widgets.Button(
            icon="fa-refresh", description="Train", layout=Layout(width="100px")
        )
        add_class_label = widgets.Label(value="Add class:")
        new_class_text = widgets.Text(layout=Layout(width="200px"))
        plus_btn = widgets.Button(icon="fa-plus", layout=Layout(width="35px"))

        button_row_list = [left_arrow_btn, progress_label, right_arrow_btn, save_btn]

        # If model arg is empty hide predict and train buttons
        if yolo_model != None:
            button_row_list = button_row_list + [predict_btn]  # train_btn

        button_row = widgets.HBox(button_row_list)
        current_img_details_row = widgets.HBox([current_img_name_label])

        bottom_row = widgets.HBox([add_class_label, new_class_text, plus_btn])

        pylabler = widgets.VBox(
            [current_img_details_row, button_row, w_bbox, bottom_row]
        )

        save_btn.on_click(on_submit)
        left_arrow_btn.on_click(on_back)
        right_arrow_btn.on_click(on_next)
        predict_btn.on_click(on_predict)
        plus_btn.on_click(on_add_class)
        new_class_text.on_submit(on_add_class)

        # Returning the container will show the widgets
        return pylabler
